#include "Owned_string.hxx"
#include "string_view.hxx"

#include <algorithm>
#include <cstdlib>
#include <cstring>

static size_t cap_for(size_t size)
{
    return size? size + 1 : size;
}

// Constructs an empty string of the given capacity. This serves
// and the default constructor as well.
Owned_string::Owned_string(size_t capacity)
        : size_(0)
        , capacity_(capacity)
        , data_(capacity? new char[capacity] : nullptr)
{ }

// Private constructor delegates to allocating constructor above,
// then copies if there's anything to copy.
Owned_string::Owned_string(size_t size, size_t capacity, const char* data)
        : Owned_string(capacity)
{
    if (size)
        prereserved_append_({data, size});
}

Owned_string::Owned_string(const char* begin, const char* end)
        : Owned_string(string_view(begin, end))
{ }

Owned_string::Owned_string(string_view sv)
        : Owned_string(sv.size(), cap_for(sv.size()), sv.begin())
{ }

Owned_string::Owned_string(Owned_string const& that)
        : Owned_string(that.size(), cap_for(that.size()), that.begin())
{ }

Owned_string::Owned_string(Owned_string&& that) noexcept
        : size_(that.size_)
        , capacity_(that.capacity_)
        , data_(that.data_)
{
    that.set_empty_();
}

/*
 * Destructor.
 */

Owned_string::~Owned_string()
{
    delete [] data_;
}

/*
 * Assignment “operators”.
 */

Owned_string& Owned_string::operator=(Owned_string const& that)
{
    if (this != &that) {
        clear();
        *this += that;
    }

    return *this;
}

Owned_string& Owned_string::operator=(Owned_string&& that) noexcept
{
    Owned_string temp(std::move(that));
    swap(temp);
    return *this;
}

/*
 * Other `String` operations.
 */

bool Owned_string::empty() const
{
    return size_ == 0;
}

size_t Owned_string::size() const
{
    return size_;
}

char Owned_string::operator[](size_t index) const
{
    return data_[index];
}

char& Owned_string::operator[](size_t index)
{
    return data_[index];
}

void Owned_string::push_back(char c)
{
    // Make sure we have room for at least one more character:
    reserve(1);
    data_[size_++] = c;
    data_[size_] = '\0';
}

void Owned_string::pop_back()
{
    data_[--size_] = '\0';
}

void Owned_string::reserve(size_t additional_cap)
{
    ensure_capacity_(cap_for(size_ + additional_cap));
}

void Owned_string::clear()
{
    if (data_) {
        size_  = 0;
        *end() = '\0';
    }
}

void Owned_string::swap(Owned_string& that)
{
    std::swap(size_,     that.size_);
    std::swap(capacity_, that.capacity_);
    std::swap(data_,     that.data_);
}

Owned_string& Owned_string::operator+=(string_view that)
{
    reserve(that.size());
    prereserved_append_(that);
    return *this;
}

const char* Owned_string::c_str() const
{
    return data_? data_ : "";
}

Owned_string::operator string_view() const
{
    if (empty())
        return {};
    else
        return {begin(), end()};
}

const char* Owned_string::begin() const
{
    return data_;
}

char* Owned_string::begin()
{
    return data_;
}

const char* Owned_string::end() const
{
    return data_ + size_;
}

char* Owned_string::end()
{
    return data_ + size_;
}

char* Owned_string::release() noexcept
{
    char* result = data_;
    set_empty_();
    return result;
}

// Clears this `String` to be the empty string with no allocated memory.
// Note that if `data_` points to a live object then calling this
// function could leak it.
void Owned_string::set_empty_() noexcept
{
    size_ = capacity_ = 0;
    data_ = nullptr;
}

// Makes sure that this string has at least capacity `min_cap`, growing
// it if necessary.
void Owned_string::ensure_capacity_(size_t min_cap)
{
    if (capacity_ < min_cap) {
        Owned_string temp(size_,
                          std::max(min_cap, 2 * capacity_),
                          data_);
        swap(temp);
    }
}

void Owned_string::prereserved_append_(string_view sv)
{
    std::copy(sv.begin(), sv.end(), end());
    size_ += sv.size();
    *end() = '\0';
}

Owned_string Owned_string::from_owned_parts(char* data, std::size_t size) noexcept
{
    Owned_string result;

    if (data) {
        result.size_     = size;
        result.capacity_ = cap_for(size);
        result.data_     = data;
    }

    return result;
}

Owned_string operator+(string_view a, string_view b)
{
    Owned_string result;
    result.reserve(a.size() + b.size());
    result.prereserved_append_(a);
    result.prereserved_append_(b);
    return result;
}

Owned_string operator+(Owned_string&& a, string_view b)
{
    a += b;
    return std::move(a);
}

